// File: dataset.cc
//
// Time-stamp: <2008-07-17 15:31:33 haog>
// Author: Guohua Hao <haog@eecs.oregonstate.edu>
//
// Copyright (C) 2008 Guohua Hao
// All Rights Reserved.
//
// Description: Implement a sparse dataset
//
//////////////////////////////////////////////////////////////////////

#include <cassert>
#include <algorithm>
#include "dataset.h"
#include "rt_main.h"

namespace treecrf {

int Example::GetFeatureValue(int index) const {
  assert(index >= 0);
  if (features_.empty()) return 0;
  int low = 0;
  int high = (int) features_.size() - 1;
  int middle_pos;
  do {
    middle_pos = (low + high) / 2;
    if (index < features_[middle_pos])
      high = middle_pos - 1;
    else
      low = middle_pos + 1;
  } while ((index != features_[middle_pos]) && (low <= high));
  if (features_[middle_pos] == index)
    return 1;
  else
    return 0;
}

void Example::SetFeatureValue(int index, int value) {
  assert(index >= 0);
  if (value == 0) {
    // Ensure that index is not present, remove if it is present
    int num_features = (int) features_.size();
    for (int key = 0; key < num_features; ++key) {
      if (features_[key] == index) {
        for (int j = key + 1; j < num_features; ++j)
          features_[j-1] = features_[j];
        features_.pop_back(); // shrink the vector
        return;
      }
    }
    return;
  } else {
    // Ensure that index is present, add if it is not present
    if (GetFeatureValue(index))
      return;
    else
      AddFeature(index);
  }
}

void Example::AddFeature(int index) {
  assert(index >= 0);
  int key = (int) features_.size();
  features_.push_back(0); // dummy value, get one extra slot in the vector
  // ripple larger elements up one slot
  while ((key > 0) && (index < features_[key - 1])) {
    features_[key] = features_[key - 1];
    key--;
  }
  // insert index in the proper spot
  features_[key] = index;
}

// The last params.npad * NClasses features encode the recurrent features.
// The formula for a second-order model looks like this
// (y_{t-1} * NClasses + y_{t-2}) * NCacheIDs + cacheID
// In the sparse representation, the recurrent features will be the last npad
// features in the feature vector.
int Example::CacheIndex() const {
  int offset = InputFeatures;
  int index = 0;
  int num_true = (int)features_.size();
  for (int i = num_true - params.npad; i < num_true; ++i) {
    index = NClasses * index + (features_[i] - offset);
    offset += NClasses;
  }
  index = index * NCacheIDs + cache_index_;
  assert(index >= 0);
  return index;
}

void DataSet::DeleteExamples() {
  for (ExamplesVecIterator itr = examples_.begin(); itr != examples_.end(); itr++)
    delete (*itr);
  examples_.clear();
}

void DataSet::Split(int feature, int value,
                    DataSet * true_child, DataSet * false_child) const {
  true_child->num_features_ = num_features_;
  true_child->num_classes_ = num_classes_;
  false_child->num_features_ = num_features_;
  false_child->num_classes_ = num_classes_;

  for (ExamplesVecIteratorConst itr = examples_.begin();
       itr != examples_.end(); itr++) {
    if ((*itr)->GetFeatureValue(feature) == value)
      true_child->AddExample(*itr);
    else
      false_child->AddExample(*itr);
  }
}

void DataSet::ClassFrequencies(std::vector<int> * class_counts) const {
  std::fill(class_counts->begin(), class_counts->end(), 0);
  for (ExamplesVecIteratorConst itr = examples_.begin();
       itr != examples_.end(); itr++)
    (*class_counts)[(*itr)->category()]++;
}

int DataSet::IsPure(const std::vector<int> & class_counts) const {
  assert((int)class_counts.size() == num_classes_);
  int nonzero_classes = 0;
  for (int iclass = 0; iclass < num_classes_; ++iclass)
    if (class_counts[iclass] > 0) nonzero_classes++;

  return (nonzero_classes < 2);
}

int DataSet::MajorityClass(const std::vector<int> & class_counts) const {
  assert((int)class_counts.size() == num_classes_);
  int best_counts = 0;
  int best_class = -1;
  for (int iclass = 0; iclass < num_classes_; ++iclass) {
    if (class_counts[iclass] > best_counts) {
      best_counts = class_counts[iclass];
      best_class = iclass;
    }
  }
  return best_class;
}

} // namespace treecrf

//////////////////////////////////////////////////////////////////////
