// File: dataset.h
//
// Time-stamp: <2008-07-17 15:32:01 haog>
// Author: Guohua Hao <haog@eecs.oregonstate.edu>
//
// Copyright (C) 2008 Guohua Hao
// All Rights Reserved.
//
// Description: This is a version of the DataSet class that supports
// sparse boolean features defined as non-negative integers. The integers
// are stored as variable-length vectors. The vectors are kept sorted
// so that lookup can be done with binary search.
//
//////////////////////////////////////////////////////////////////////

#ifndef _DATASET_H_
#define _DATASET_H_

#include <cassert>
#include <vector>

namespace treecrf {

class Example {
 public:
  // Constructors
  // Create a null example
  Example(): features_(0), category_(0), target_(0.0), skip_(false),
             cache_index_(-1) {};

  // Specify the number of features
  explicit Example(int number_elements):
    features_(number_elements), category_(0), target_(0.0), skip_(false),
    cache_index_(-1) {};

  // Specify the number of features and initial values. Default category is 0
  Example(int number_elements, int init_value):
    features_(number_elements), category_(0), target_(0.0), skip_(false),
    cache_index_(-1) {
    assert(init_value == 0); // init_value must be zero for sparse datasets
  };

  // Copy an example
  Example(const Example & source): features_(source.features_),
                                   category_(source.category_),
                                   target_(source.target_),
                                   skip_(source.skip_),
                                   cache_index_(source.cache_index_) {};

  // Return value of feature index. Binary search version.
  int GetFeatureValue(int feature_index) const;

  // Set the feature index to value
  void SetFeatureValue(int index, int value);

  // Add new true feature value to feature vector.
  // assume that this feature is not in the feature vector yet.
  void AddFeature(int index);

  // Compute an index into the cache based on the recurrent features.
  int CacheIndex() const;

  // Get the number of true features in the example
  int GetSize() const {return (int)features_.size(); }

  // Get the feature at location 'loc' in the vector 'features_'
  int GetFeatureAt(int loc) const { return features_[loc]; };

  // Set the feature at location 'loc' with feature index 'index'
  void SetFeatureAt(int loc, int index) { features_[loc] = index; };

  // Getters and Setters
  int category() const { return category_; };
  double target() const {return target_; };
  bool skip() const {return skip_; };
  void set_category(int label) { category_ = label; };
  void set_target(double target) { target_ = target; };
  void set_skip(bool skip) { skip_ = skip; };
  void set_cache_index(int cache_index) { cache_index_ = cache_index; };

 private:
  // Data Fields
  std::vector<int> features_; // indices of the true features
  int category_;
  double target_; // auxiliary target value during gradient boosting
  bool skip_; // true if we should skip this instance
  int cache_index_; // used for caching. -1 means not cached
};

class DataSet {
 public:
  typedef std::vector<Example *>::iterator ExamplesVecIterator;
  typedef std::vector<Example *>::const_iterator ExamplesVecIteratorConst;

  // constructors and destructor
  // Null dataset
  DataSet(): examples_(0), num_features_(0), num_classes_(0) {};

  DataSet(int num_features, int num_classes): examples_(0),
                                              num_features_(num_features),
                                              num_classes_(num_classes) {};

  // Copy constructor.  This does a shallow copy of the examples via
  // the vector copy constructor.
  DataSet(const DataSet & other): examples_(other.examples_),
                                  num_features_(other.num_features_),
                                  num_classes_(other.num_classes_) {};

  // Reclaim the storage occupied by the examples in the dataset.
  void DeleteExamples();

  // Add an example to a dataset
  void AddExample(Example * ex) { examples_.push_back(ex); };

  // Operations for supporting decision tree construction. Create two new
  // datasets, where we have split the current dataset according to feature
  // 'feature' and value 'value'.
  void Split(int feature, int value,
             DataSet * true_child, DataSet * false_child) const;

  // Compute the frequency of each class label in the dataset;
  void ClassFrequencies(std::vector<int> * counts) const;

  // Return true if all examples in this dataset belong to one class
  int IsPure(const std::vector<int> & freq) const;

  // Return the majority class in the dataset
  int MajorityClass(const std::vector<int> & freq) const;

  Example * GetExample(int index) const { return examples_[index]; }
  int GetNumExamples() const { return (int)examples_.size(); };

  // Getters and Setters
  int num_features() const { return num_features_; };
  int num_classes() const { return num_classes_; };
  void set_num_features(int num_features) { num_features_ = num_features; };
  void set_num_classes(int num_classes) { num_classes_ = num_classes; };

 private:
  // Data Fields
  std::vector<Example *> examples_;
  int num_features_; // This is the maximum number of possible features
  int num_classes_; // this value does not include the off-the-end class '-1'
};

} // namespace treecrf

#endif // _DATASET_H_
//////////////////////////////////////////////////////////////////////
