// File: forward_backward.h
//
// Time-stamp: <2008-07-17 18:13:41 haog>
// Author: Guohua Hao <haog@eecs.oregonstate.edu>
//
// Copyright (C) 2008 Guohua Hao
// All Rights Reserved.
//
// Description: ForwardBackwardEval class, which is a forward_backward
// algorithm wrapper.
//
//////////////////////////////////////////////////////////////////////

#ifndef _FORWARD_BACKWARD_H_
#define _FORWARD_BACKWARD_H_

#include <list>
#include <queue>
#include <vector>
#include "viterbi_node.h"

namespace treecrf {

class DataSet;
class TreeCRF;

class ForwardBackwardEval {
 public:
  typedef std::vector<ViterbiNode *> VNodePtrVec;
  typedef VNodePtrVec::iterator VecIter;
  typedef std::vector<VNodePtrVec *> VNodePtrMatrix;
  typedef VNodePtrMatrix::iterator MatrixIter;
  typedef std::list<ViterbiNode *> VNodePtrList;

  // Constructors
  ForwardBackwardEval(): k_val_(1), beam_breadth_(10) {};

  ForwardBackwardEval(int k_val, int beam_breadth):
    k_val_(k_val), beam_breadth_(beam_breadth) {};

  // Run the forward backward algorithm to generate the trellis for a given
  // sequence 'dataset' using 'crf'
  void ForwardBackward(const DataSet & dataset,
                       const TreeCRF & crf,
                       VNodePtrMatrix * trellis);

  // Prediction the label sequence of a given sequence dataset with crf
  void Predict(const DataSet & dataset,
               const TreeCRF & crf,
               int * predicted_classes);

  // Reclaim the memories for all nodes in the trellis and the trellis itself.
  // Remember to call this function after you are done with the trellis.
  void DeleteTrellis(VNodePtrMatrix * trellis);

 private:
  // Generate the next heap from the parents
  void GrowNextHeap(VNodePtrList & parents, int num_classes);

  // Builds a sublist of parents, where each node has the same possible range
  // for their children
  void BuildSubParents(VNodePtrList & parents, VNodePtrList & sub_parents);

  // Potentially put the current node into the PQ during the beam search
  void UpdateHeap(ViterbiNode * current);

  // build the next column of the trellis based on the nodes
  // in the PQ, i.e., the beam
  void BuildNodeMatrix(VNodePtrVec * current_column);

  // Assign each node in the current column their probabilities
  void BuildParents(VNodePtrVec * current_column,
                    Example * ex,
                    const TreeCRF & crf,
                    int num_features,
                    int num_classes,
                    VNodePtrList & parents);

  // Backward pass
  void BetaStep(VNodePtrMatrix * trellis);

  // Data Fields
  int k_val_; // the order of markov chain
  int beam_breadth_; // the breadth of beam search
  std::priority_queue<ViterbiNode*, std::vector<ViterbiNode *>, ViterbiNodeComparator> best_nodes_;
};

} // namespace treecrf

#endif  // _FORWARD_BACKWARD_H_
//////////////////////////////////////////////////////////////////////
