// File: parameters.h
//
// Time-stamp: <2008-07-18 19:13:09 haog>
// Author: Guohua Hao <haog@eecs.oregonstate.edu>
//
// Copyright (C) 2008 Guohua Hao
// All Rights Reserved.
//
// Description: A class for reading and printing the parameters
// controlling the TreeCRF program.
//
//////////////////////////////////////////////////////////////////////

#ifndef _PARAMETERS_H_
#define _PARAMETERS_H_

#include <iostream>
#include <string>

namespace treecrf {

class Parameters {
 public:
  // constructors
  // default initialization
  Parameters();

  // read parameters from file
  Parameters(char *);

  // Display the parameters used in the TreeCRF program
  void Display(std::ostream & str);

  // Data Fields

  // Critical parameters
  // the number of training iterations to run
  int iterations;

  // max leaves for each crf regression tree
  int maxLeaves;

  // order of Markov random field (>= 1)
  int npad;

  // 'testInterval' sets how often the crf evaluates the testing dataset,
  // ex: if (iteration % testInterval == 0) doTest
  int testInterval;

  // 'saveInterval' sets how often the crf saves the accumulated trees
  int saveInterval;

  // file stem for prediction outputs
  std::string filestem;

  // the beam width for fb training and evaluation
  int fbWidth;

  // the beam width for viterbi evaluation
  int viterbiWidth;

  // Selects the classifcation method. Both set to zero selects viterbi.
  // This is pretty dumb. I'll replace it with a single variable allowing
  // tree values (rsw, fb, viterbi).
  int predictRSW;
  int predictFB;

  // Toggles printing out the cpu time taken for each iteration
  int printTime;

  // -------------------------------------------------------------------
  // Parameters related to avoiding overfitting
  // -------------------------------------------------------------------

  // 'sampleRate' sets the ratio of training sequences randomly
  // selected for each training iteration.  The range is 0 - 1.
  double sampleRate;

  // 'scale' is a factor multiplying every leaf output value in the crf.
  // This provides a very crude way to control the size of gradient steps.
  double scale;

  // 'trim' enables (trim=1) or disables (trim=0) influence trimming.
  // 'trim_val' is the threshold used when trimming examples.
  int trim;
  double trim_val;

  // 'treeFeatureRate' allows the crf to train using randomized trees.
  // 'treeFeatureRate' sets the ratio of the total atomic features that
  // are considered when growing split nodes in the regression trees.
  double treeFeatureRate;

  // 'iterCutoff' is a threshold that sets how many iterations of plain
  // boosting regression trees are grown before CRF training begins. It
  // should normally be set to 0. 'initMaxLeaves' sets the tree size of
  // these initial regression trees.
  int iterCutoff;
  int initMaxLeaves;

  // the weight of a quadratic penalty over the leaf values in growing
  // regression trees. This shrinks the leaf values towards zero.
  double shrinkage;

  // -------------------------------------------------------------------
  // Other parameters...
  // -------------------------------------------------------------------

  // The maximum examples allowed in a data set.
  int maxSetSize;

  // Toggles information about viterbi and fb on and off.  This was used for
  // bug checking.
  int printViterbi;
  int printFB;

  // Toggles normalization while computing fb or viterbi sweeps.  This must
  // be enables for data with long sequences!
  int normalize;

  // Toggles computing the beta (backward) pass of the fb algorithm. This
  // should normally be on (betaPass=1).
  int betaPass;

  // Toggles how the most likely class in the fb classification is selected.
  // The 'correct' way to calucate the classification is to have this enabled.
  // The parameter could be removed.
  int sum_fb;

 private:
  void Read(char * filename);
  void skipComments(std::istream & str);
};

} // namespace treecrf

#endif // _PARAMETERS_H_
//////////////////////////////////////////////////////////////////////
