// File: potential_function.cc
//
// Time-stamp: <2008-07-18 11:27:50 haog>
// Author: Guohua Hao <haog@eecs.oregonstate.edu>
//
// Copyright (C) 2008 Guohua Hao
// All Rights Reserved.
//
// Description: implementation of the Regression Tree Potential Function
//
//////////////////////////////////////////////////////////////////////

#include <cmath>
#include <string>
#include <list>
#include "dataset.h"
#include "potential_function.h"
#include "regression_tree.h"
#include "rt_main.h"

namespace treecrf {

RTPotentialFunction::~RTPotentialFunction() {
  std::list<RegressionTree *>::iterator itr;
  for (itr = trees_.begin(); itr != trees_.end(); itr++)
    delete *itr;
}

double RTPotentialFunction::Score(const Example & ex) {
  // Cache lookup version
  // New regression trees are added to the back of the trees list.
  // If the cached value is one generation old, then we just evaluate
  // that tree.  This will be the common case.  Otherwise, we evaluate them all.
  const int check_cache = 0;
  double full_score = 0.0;
  if (check_cache) {
    full_score = RawScore(ex);
  }

  int index = ex.CacheIndex();
  int cached_generation = generation_[index];
  double score = 0.0;
  if (cached_generation == current_generation_) {
    CachePerfectHits++;
    if (check_cache && fabs(cache_[index] - full_score) > 1e-6) {
      std::cout << "Cache gives wrong value: " << cache_[index]
                << " fullScore = " << full_score << std::endl;
      return full_score;
    }
    return cache_[index];
  } else if (cached_generation == current_generation_ - 1) {
    CacheOneHits++;
    score = cache_[index] + params.scale * trees_.back()->PredictExample(ex);
  } else {
    CacheMisses++;
    score = RawScore(ex);
  }

  // Update the cached values and generation
  cache_[index] = score;
  generation_[index] = current_generation_;
  if (check_cache && fabs(score - full_score) > 1e-6) {
    std::cout << "Cache gives wrong value: " << score
              << " fullScore = " << full_score << std::endl;
    return full_score;
  }
  return score;
}

double RTPotentialFunction::RawScore(const Example & ex) const {
  double score = 0.0;
  std::list<RegressionTree *>::const_iterator itr;
  for (itr = trees_.begin(); itr != trees_.end(); itr++) {
    score += params.scale * (*itr)->PredictExample(ex);
  }
  return score;
}

void RTPotentialFunction::Train(const DataSet & dataset, int num_leaves) {
  RegressionTree * rt = new RegressionTree(num_leaves);
  rt->MakeRegressionTree(dataset);
  trees_.push_back(rt);
  current_generation_++;
}

// Output format:
// RTPotentialFunction length tree_1 ... tree_n
void RTPotentialFunction::Print(std::ostream & str) const {
  str << "RTPotentialFunction " << (int)trees_.size();
  std::list<RegressionTree *>::const_iterator itr;
  for (itr = trees_.begin(); itr != trees_.end(); itr++)
    str << " " << **itr;
  str << " ";
}

void RTPotentialFunction::Read(std::istream & str) {
  std::string buf;
  str >> buf;
  if (buf != std::string("RTPotentialFunction")) {
    std::cerr << "Found `" << buf
              << "' where RTPotentialFunction was expected." << std::endl;
    std::abort();
  }
  int ntrees = 0;
  str >> ntrees;
  for (int itree = 0; itree < ntrees; ++itree) {
    RegressionTree * rt = new RegressionTree();
    str >> *rt;
    trees_.push_back(rt);
  }
  // update the current_generation_ to the current number of trees.
  current_generation_ = ntrees;
}

void operator << (std::ostream & str, const RTPotentialFunction & pf) {
  return pf.Print(str);
}

void operator >> (std::istream & str, RTPotentialFunction & pf) {
  return pf.Read(str);
}

} // namespace treecrf

//////////////////////////////////////////////////////////////////////
