// File: regression_tree.cc
//
// Time-stamp: <2008-07-24 17:43:51 haog>
// Author: Guohua Hao <haog@eecs.oregonstate.edu>
//
// Copyright (C) 2008 Guohua Hao
// All Rights Reserved.
//
// Description: build a regression tree from a given dataset.
//
//////////////////////////////////////////////////////////////////////

#include <cassert>
#include <iostream>
#include "dataset.h"
#include "regression_tree.h"
#include "rt_main.h"

namespace treecrf {

void RegressionTreeNode::InitRoot() {
  int nexamples = dataset_->GetNumExamples();
  if (nexamples > 0) {
    // compute MSE at root
    double sumX = 0.0;
    for (int iexample = 0; iexample < nexamples; ++iexample) {
      Example * ex = dataset_->GetExample(iexample);
      if (!ex->skip()) sumX += ex->target();
    }
    output_ = sumX / (params.shrinkage + (double)nexamples);
    if (0) std::cout << "Root mean = " << output_ << std::endl;
  } else {
    // if no training examples, force a leaf
    output_ = 0;
  }
  is_leaf_ = true;
  true_child_ = NULL;
  false_child_ = NULL;
}

bool RegressionTreeNode::FindBestSplit() {
  assert(is_leaf_);

  // statistics for the current node
  int count = dataset_->GetNumExamples();
  double sumX = 0.0;

  // Statistics for the true child node, one for each possible split.
  // We are inefficient here, since we only need to consider features
  // that are actually true for at least one of the examples.  But we
  // don't have that information handy.
  int nfeatures = dataset_->num_features();
  std::vector<double> trueSumX(nfeatures, 0);
  std::vector<int> trueCount(nfeatures, 0);

  // It turns out that we can compute the gain in penalized squared
  // error using only sumX statistics, which is surprising. But the
  // sum x^2 quantities cancel.
  // The formulas involve subtractions, which is risky, but because
  // our values are in the range [-1,+1], experiments show that it is
  // safe.
  // We will take an added risky step of computing sumx for the false
  // branch by subtracting:  sumX - trueSumX

  // For each example reaching this node compute the sufficient statistics
  for (int iexample = 0; iexample < count; ++iexample) {
    Example * ex = dataset_->GetExample(iexample);
    double target = ex->target();
    sumX += target;

    // For each true feature of this example increment the sufficient statistics
    // for the "true" branch to evaluate splitting on this feature.
    int num_true_features = ex->GetSize();
    for (int j = 0; j < num_true_features; ++j) {
      int index = ex->GetFeatureAt(j);
      trueSumX[index] += target;
      trueCount[index]++;
    }
  }

  double parentMean = output_;

  // Now loop through features and compute the reduction
  // in (penalized) squared error if we split on this feature.
  // Remember the feature that gives the biggest such reduction.
  //
  // The formula for the gain is
  // (2*lambda + trueCount) * trueMean^2 + (2*lambda + falseCount) * falseMean^2
  //   - (2*lambda + count) * parentmean^2
  // This is a dangerous subtraction.
  // Note also that our penalty depends on the absolute value of the
  // mean which would be a bit odd in ordinary regression trees, but
  // which makes good sense for this potential function trees.

  // Initialize the information in the leaf
  split_feature_ = -1;
  split_value_ = 1;  // threshold, not used yet
  split_score_ = 0.0;

  for (int ifeature = 0; ifeature < nfeatures; ++ifeature) {
    double tc = (double) (trueCount[ifeature]);
    double falseCount = ((double)(count - tc));

    // If either side would have fewer than 2 examples, skip this feature.
    if ((tc < 2) || (falseCount < 2)) {
      // std::cout << "min data pruning feature " << ifeature << std::endl;
      continue;
    }

    // compute penalized means
    double trueMean = trueSumX[ifeature] / (params.shrinkage + tc);
    double falseMean = (sumX - trueSumX[ifeature]) / (params.shrinkage + falseCount);

    // The gain is actually -(reduction in squared error) because our priority
    // queue minimizes.
    double gain = (2*params.shrinkage + count) * parentMean * parentMean -
      ( (2*params.shrinkage + tc) * trueMean * trueMean +
	(2*params.shrinkage + falseCount) * falseMean * falseMean );

    if (gain < split_score_) {
      // new best split
      split_feature_ = ifeature;
      split_value_ = 1;
      split_score_ = gain;
      true_child_output_ = trueMean;
      false_child_output_ = falseMean;
    }
  }

  return (split_feature_ != -1);
}

void RegressionTreeNode::Split() {
  assert(split_feature_ >= 0);
  assert(is_leaf_);

  DataSet * true_dataset = new DataSet;
  DataSet * false_dataset = new DataSet;
  dataset_->Split(split_feature_, split_value_, true_dataset, false_dataset);

  true_child_ = new RegressionTreeNode(true_dataset, true, true_child_output_);
  false_child_ = new RegressionTreeNode(false_dataset, true, false_child_output_);

  // Reclaim the memory and make it nonleaf
  delete dataset_;
  dataset_ = 0;
  is_leaf_ = false;
}

void RegressionTreeNode::ClearData() {
  if (is_leaf_) {
    delete dataset_;
    dataset_ = 0;
    return;
  } else {
    true_child_->ClearData();
    false_child_->ClearData();
  }
}

double RegressionTreeNode::PredictExample(const Example & ex) const {
  if (is_leaf_) {
    return output_;
  } else {
    if (ex.GetFeatureValue(split_feature_) == split_value_)
      return true_child_->PredictExample(ex);
    else
      return false_child_->PredictExample(ex);
  }
}

void RegressionTreeNode::ShowTree(int nindent) const {
  if (is_leaf_) {
    Indent(nindent);
    std::cout << "output = " << output_ << std::endl;
  } else {
    Indent(nindent);
    std::cout << "feature " << split_feature_ <<  " == " << split_value_
              << std::endl;
    true_child_->ShowTree(nindent + 2);
    Indent(nindent);
    std::cout << "feature " << split_feature_ <<  " != " << split_value_
              << std::endl;
    false_child_->ShowTree(nindent + 2);
  }
}

void RegressionTreeNode::Print(std::ostream & str) const {
  str << is_leaf_ << " ";
  if (is_leaf_) {
    str << output_ << std::endl;
  } else {
    str << split_feature_ << " " << split_value_ << " ";
    true_child_->Print(str);
    false_child_->Print(str);
  }
}

void RegressionTreeNode::Read(std::istream & str) {
  str >> is_leaf_;
  if (is_leaf_) {
    str >> output_;
  } else {
    str >> split_feature_ >> split_value_;
    true_child_ = new RegressionTreeNode;
    true_child_->Read(str);
    false_child_ = new RegressionTreeNode;
    false_child_->Read(str);
  }
}

void RegressionTreeNode::Indent(int nindent) const {
  for (int i = 0; i < nindent; ++i)
    std::cout << " ";
}

void RegressionTree::MakeRegressionTree(const DataSet & dataset) {
  DataSet * ds = new DataSet(dataset);
  root_ = new RegressionTreeNode(ds);
  root_->InitRoot();
  int nleaves = 1;
  // Now add splits to the tree until max treesize is reached
  if (dataset.GetNumExamples() > 0) {
    if (root_->FindBestSplit()) next_splits_.push(root_);
    while ((nleaves < max_leaves_) && !next_splits_.empty()) {
      SplitBestLeaf();
      nleaves++;
    }
  }
  // Traverse the tree to delete all of the datasets associated with leaf nodes
  // and reclaim memory.
  root_->ClearData();
  // clear the priority queue
  ClearQueue();
}

double RegressionTree::PredictExample(const Example & ex) const {
  assert(root_ != 0);
  return root_->PredictExample(ex);
}

void RegressionTree::ShowTree(int nindent) const {
  assert(root_ != 0);
  root_->ShowTree(nindent);
}

void RegressionTree::Print(std::ostream & str) const {
  assert(root_ != 0);
  root_->Print(str);
}

void RegressionTree::Read(std::istream & str) {
  root_ = new RegressionTreeNode;
  root_->Read(str);
}

void RegressionTree::DeleteTree(RegressionTreeNode * root) {
  if (!root->is_leaf()) {
    DeleteTree(root->true_child());
    DeleteTree(root->false_child());
  }
  delete root;
  root = 0;
}

void RegressionTree::SplitBestLeaf() {
  // parent is the leaf to split
  RegressionTreeNode * parent = next_splits_.top();
  next_splits_.pop();
  parent->Split();   // Split the parent node into two children nodes

  RegressionTreeNode * true_child = parent->true_child();
  if (true_child->FindBestSplit()) next_splits_.push(true_child);
  RegressionTreeNode * false_child = parent->false_child();
  if (false_child->FindBestSplit()) next_splits_.push(false_child);
}

void RegressionTree::ClearQueue() {
  while (!next_splits_.empty())
    next_splits_.pop();
}

void operator << (std::ostream& str, const RegressionTree & tree) {
  tree.Print(str);
}

void operator >> (std::istream& str, RegressionTree & tree) {
  tree.Read(str);
}

} // namespace treecrf

//////////////////////////////////////////////////////////////////////
