// File: regression_tree.h
//
// Time-stamp: <2008-07-18 11:53:41 haog>
// Author: Guohua Hao <haog@eecs.oregonstate.edu>
//
// Copyright (C) 2008 Guohua Hao
// All Rights Reserved.
//
// Description: RegressionTreeNode class and RegressionTree class.
//
//////////////////////////////////////////////////////////////////////

#ifndef _REGRESSION_TREE_H_
#define _REGRESSION_TREE_H_

#include <iostream>
#include <queue>

namespace treecrf {

class DataSet;
class Example;

// Regression tree node classe
class RegressionTreeNode {
 public:
  // Constructors
  // for read
  RegressionTreeNode(): dataset_(0) {};
  // Top level
  explicit RegressionTreeNode(DataSet * dataset): dataset_(dataset) {};
  // For split
  RegressionTreeNode(DataSet * dataset, bool is_leaf, double output):
    dataset_(dataset), is_leaf_(is_leaf), output_(output) {};

  // Init this node as the root node
  void InitRoot();

  // Determine the best attribute to split on at the current node.
  // Return true if the best attribute exists, return false otherwise.
  bool FindBestSplit();

  // Split this node into two children nodes
  void Split();

  // Recursively reclaim all of the data sets stored in the tree.
  // We need to keep data in the leaves until the tree stops being split.
  // So we only need to reclaim the data sets associated with leaf nodes.
  void ClearData();

  // Evaluate the regression tree over an exammple, only called on the root node.
  double PredictExample(const Example & ex) const;

  // Print out the tree
  void ShowTree(int nindent) const;

  // Getters
  bool is_leaf() const { return is_leaf_; };
  double split_score() const { return split_score_; };
  RegressionTreeNode * true_child() const { return true_child_; };
  RegressionTreeNode * false_child() const { return false_child_; };

  // IO operation
  void Print(std::ostream & str) const;
  void Read(std::istream & str);

 private:
  // Indent n characters by putting n space
  void Indent(int nindent) const;

  // Data Fields
  DataSet * dataset_; // temporarily store data, so that we can split leaves
  bool is_leaf_;
  double output_; // predicted real value at this node

  int split_feature_;
  int split_value_;
  double split_score_; // Reduction in squared error compared to parent (negative is good)

  double true_child_output_;
  double false_child_output_;

  RegressionTreeNode * true_child_;
  RegressionTreeNode * false_child_;
};

// Class for compare the split_score of two regression tree nodes, which are the
// elements of a priority queue
class RegressionTreeNodeComparator {
 public:
  bool operator() (RegressionTreeNode * a, RegressionTreeNode * b) {
    if (a->split_score() > b->split_score()) return true;
    return false;
  }
};

class RegressionTree {
 public:
  // Constructors
  // Default constructor
  RegressionTree(): root_(0), max_leaves_(0) {};
  // Specify the max number of leaves
  explicit RegressionTree(int max_leaves): root_(0), max_leaves_(max_leaves) {};

  ~RegressionTree() { DeleteTree(root_); };

  // Build a regression tree based on the provided dataset
  void MakeRegressionTree(const DataSet & dataset);

  // Predict the output value for an given example
  double PredictExample(const Example & ex) const;

  // print out the tree
  void ShowTree(int nindent) const;

  // IO operation
  void Print(std::ostream & str) const;
  void Read(std::istream & str);

 private:
  // Delete the regression tree rooted at 'root'
  void DeleteTree(RegressionTreeNode * root);

  // Pop best leaf from next_splits_ priority queue, split it, and push children
  // nodes into the pq if possible
  void SplitBestLeaf();

  // Make the PQ empty;
  void ClearQueue();

  // Data fields
  // The root of the regression tree
  RegressionTreeNode * root_;
  // the maximum number of leaves in the regression tree
  int max_leaves_;
  // priority queue for best-first tree growing.
  std::priority_queue<RegressionTreeNode *, std::vector<RegressionTreeNode *>, RegressionTreeNodeComparator> next_splits_;
};

void operator << (std::ostream & str, const RegressionTree & tree);
void operator >> (std::istream & str, RegressionTree & tree);

} // namespace treecrf

#endif  // _REGRESSION_TREE_H_
//////////////////////////////////////////////////////////////////////
