// File: rt-main.cc
//
// Time-stamp: <2008-07-18 18:36:41 haog>
// Author: Guohua Hao <haog@eecs.oregonstate.edu>
//
// Copyright (C) 2008 Guohua Hao
// All Rights Reserved.
//
// Description: main routine
//
//////////////////////////////////////////////////////////////////////

#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <fstream>
#include <iostream>
#include <vector>
#include "dataset.h"
#include "forward_backward.h"
#include "parameters.h"
#include "rsw.h"
#include "statistics.h"
#include "treecrf.h"
#include "viterbi.h"

namespace treecrf {

// Global variables
Parameters params;

int heapshrink = 0;

int CachePerfectHits = 0;
int CacheOneHits = 0;
int CacheMisses = 0;

int NCacheIDs = 0;
int CacheSize = 0;

int InputFeatures = 0;
int NClasses = 0; // includes the extra (-1) off-end class

void BuildSet(char * filename, int k_val, std::vector<DataSet *> * train_words) {
  int num_sequences, output_features;
  std::ifstream str(filename);
  str >> num_sequences;
  str >> NClasses;
  str >> InputFeatures;
  NClasses++; // Introduce a Null class for positions off the end of the
              // sequence. This will be class -1.  It is used for recurrent
              // class ids to the left of the sequence.
  output_features = NClasses * k_val;
  train_words->resize(num_sequences);

  int crnt_seq, crnt_time, crnt_feature, crnt_class, ntrue;

  // read the sequence number for the next example
  str >> crnt_seq;
  int last_seq = crnt_seq; // current sequence number in file
  int sequence_skip = 0;

  for  (int seq = 0; seq < num_sequences; ++seq) {
    DataSet * ds = new DataSet(InputFeatures + output_features, NClasses - 1);
    std::vector<int> old_states(k_val, -1);
    while (crnt_seq == last_seq) {
      str >> crnt_time;
      str >> ntrue;
      // allocate a new example and fill in its input and output features
      Example * ex = new Example;
      for (int itoken = 0; itoken < ntrue; ++itoken) {
	str >> crnt_feature;
        ex->AddFeature(crnt_feature);
      }

      // Now let x denote a feature index.
      // The features numbered starting at InputFeatures are indicator
      // variables of the classes at various positions
      int x = InputFeatures;
      // y denotes a preceding time step
      for (int y = 0; y < k_val; ++y) {
        ex->AddFeature(x + old_states[y] + 1);
        x += NClasses;
      }
      // read the class label
      str >> crnt_class;
      ex->set_category(crnt_class);
      // Assign the cache id
      ex->set_cache_index(NCacheIDs++);

      ds->AddExample(ex);

      if (k_val > 0) {
	// Slide the previous states forward one step
	for (int x = k_val - 1; x > 0; x--)
	  old_states[x] = old_states[x-1];
	old_states[0] = crnt_class;
      }
      str >> crnt_seq;
    } // while sequence is unchanged

    (*train_words)[seq] = ds;
    if (crnt_seq < last_seq) {
      std::cerr << "Error: sequence number decreased in data file!" << std::endl;
      std::abort();
    } else if (crnt_seq != last_seq + 1) {
      // we have a skip
      std::cout << "Sequence number skipped from " << last_seq << " to " << crnt_seq << std::endl;
      sequence_skip = 1;
      num_sequences -= crnt_seq - (last_seq + 1);
    }
    last_seq = crnt_seq;
  }
  // update the size of the train word set
  if (sequence_skip) train_words->resize(num_sequences);
}

void PrintUsage() {
  std::cout << "usage: rt -p params -t traindata -x testdata -s savefile -r restorefile"
            << std::endl;
  std::abort();
}

void SaveClassifier(char * name,
		    int iteration,
		    double cum_time,
		    const TreeCRF & crf) {
  char filename[256];
  std::sprintf(filename, "%s-%d", name, iteration);

  std::ofstream str(filename);
  if (!str)
    std::cerr << "Unable to open tree save file " << filename << std::endl;

  // number of iterations and total time
  str << iteration << " " << cum_time << std::endl;
  // the forest
  str << crf;
}

// reclaim the memory used to store examples.
void DeleteAllExamples(std::vector<DataSet *> & words) {
  for (std::vector<DataSet *>::iterator itr = words.begin();
       itr != words.end(); itr++) {
    (*itr)->DeleteExamples();
    delete *itr;
  }
}

} // namespace treecrf

int main(int argc, char * argv[]) {
  using namespace treecrf;

  std::srand(time(NULL));

  int optc;
  char * paramfile = 0;
  int fparamfile = 0;
  char * trainfile = 0;
  int ftrainfile = 0;
  char * testfile = 0;
  int ftestfile = 0;
  char * writetreefile = 0;
  int fwritetreefile = 0;
  char * readtreefile = 0;
  int freadtreefile = 0;

  while ((optc = getopt(argc, argv, "p:t:x:s:r:")) != -1 ) {
    switch (optc) {
    case 'p':
      paramfile = optarg;
      fparamfile = 1;
      std::cout << "Read parameters from " << paramfile << std::endl;
      break;
    case 't':
      trainfile = optarg;
      ftrainfile = 1;
      std::cout << "Read training data from " << trainfile << std::endl;
      break;
    case 'x':
      testfile = optarg;
      ftestfile = 1;
      std::cout << "Read test data from " << testfile << std::endl;
      break;
    case 's':
      writetreefile = optarg;
      fwritetreefile = 1;
      std::cout << "Save trained classifier on " << writetreefile << std::endl;
      break;
    case 'r':
      readtreefile = optarg;
      freadtreefile = 1;
      std::cout << "Read trained classifier from " << readtreefile << std::endl;
      break;
    default:
      PrintUsage();
    }
  }

  if (!fparamfile || !ftrainfile) PrintUsage();

  params = Parameters(paramfile);
  std::cout << "Parameters:\n";
  params.Display(std::cout);

  std::vector <DataSet *> train_words(params.maxSetSize);
  std::vector <DataSet *> test_words(params.maxSetSize);

  if (ftrainfile) {
    std::cout << "Loading train words..." << std::endl;
    BuildSet(trainfile, params.npad, &train_words);
    std::cout << "read " << train_words.size() << " training sequences" << std::endl;
  }

  if (ftestfile) {
    std::cout << "Loading test words..." << std::endl;
    BuildSet(testfile, params.npad, &test_words);
    std::cout << "read " << test_words.size() << " test sequences" << std::endl;
  }

  CacheSize = 1;
  for (int i = 0; i < params.npad; ++i)
    CacheSize = CacheSize * NClasses;
  CacheSize = CacheSize * NCacheIDs;
  std::cout << "CacheSize = " << CacheSize << std::endl;

  // Allocate the potential functions
  TreeCRF crf(params.npad, train_words[0]->num_classes(), CacheSize);
  ForwardBackwardEval fb_eval(params.npad, params.fbWidth);

  int iteration = 0;
  double cum_time = 0;
  if (freadtreefile) {
    std::ifstream str(readtreefile);
    str >> iteration >> cum_time >> crf;
    iteration++;
  }

  for (; iteration < params.iterations; ++iteration) {
    std::cout << std::endl << "Training Iteration " <<  iteration << "..."
              << std::endl;

    std::clock_t start_time, end_time;
    start_time = clock();
    if (iteration < params.iterCutoff)
      crf.BoostRSW(train_words);
    else
      crf.BoostML(train_words, fb_eval);
    end_time = clock();

    double iter_time = ((double)(end_time - start_time))/(double)CLOCKS_PER_SEC;
    cum_time += iter_time;

    if (params.printTime)
      std::cout << "Train time: " << iter_time << " sec." << std::endl;

    std::cout << "CachePerfectHits = " << CachePerfectHits
              << " CacheOneHits = " << CacheOneHits
              << " CacheMisses = " << CacheMisses << std::endl;

    if (((iteration % params.testInterval) == 0) && ftestfile) {
      char filename[120];
      std::sprintf(filename, "%s-%d", params.filestem.c_str(), iteration);
      std::ofstream str(filename);
      std::cout << "Test Set Results..." << std::endl;

      Statistics stats;
      stats.words = stats.wordErrors = 0;
      stats.letters = stats.letterErrors = 0;

      int k_val, nwindows;
      k_val = params.npad;

      // std::cout << "Errors: " << std::endl;
      stats.words += test_words.size();
      for (std::vector<DataSet *>::iterator itr = test_words.begin();
           itr != test_words.end(); itr++) {
	nwindows = (*itr)->GetNumExamples();
	int trueClasses[nwindows];
	int predictedClasses[nwindows];

	for (int letterIter = 0; letterIter < nwindows; ++letterIter)
	  trueClasses[letterIter] = (*itr)->GetExample(letterIter)->category();

 	if (params.predictRSW) {
          RswEval rsw_eval(params.npad);
          rsw_eval.Predict(**itr, crf, predictedClasses);
        } else if (params.predictFB) {
          fb_eval.Predict(**itr, crf, predictedClasses);
        } else {
          ViterbiEval v_eval(params.npad, params.viterbiWidth);
          v_eval.Predict(**itr, crf, predictedClasses);
        }

        stats.letters += nwindows;
	int wordError = 0;
	for (int i = 0; i < nwindows; ++i) {
	  int letterError = (predictedClasses[i] != trueClasses[i]);
	  if (letterError) stats.letterErrors++;
	  wordError |= letterError;
	  str << predictedClasses[i] << std::endl;
	}
	if (wordError) stats.wordErrors++;
      }

      stats.phonemeErrors = stats.letters;
      stats.stressErrors = stats.letters;
      stats.Show(iteration, iter_time, cum_time);
      std::cout << "heapshrink = " << heapshrink << std::endl;
    }

    if (((iteration % params.saveInterval) == 0) && fwritetreefile)
      SaveClassifier(writetreefile, iteration, cum_time, crf);
  }

  if (fwritetreefile)
    SaveClassifier(writetreefile, params.iterations - 1 , cum_time, crf);

  // reclaim the memory used to store the examples
  DeleteAllExamples(train_words);
  if (ftestfile) DeleteAllExamples(test_words);

  return 0;
}

//////////////////////////////////////////////////////////////////////
