// File: treecrf.cc
//
// Time-stamp: <2008-07-18 19:09:04 haog>
// Author: Guohua Hao <haog@eecs.oregonstate.edu>
//
// Copyright (C) 2008 Guohua Hao
// All Rights Reserved.
//
// Description: implementation of TreeCRF class
//
//////////////////////////////////////////////////////////////////////

#include <cmath>
#include <cstdlib>
#include <string>
#include <vector>
#include <algorithm>
#include "dataset.h"
#include "forward_backward.h"
#include "potential_function.h"
#include "rt_main.h"
#include "treecrf.h"

namespace treecrf {

TreeCRF::~TreeCRF() {
  std::vector<RTPotentialFunction *>::iterator itr;
  for (itr = potentials_.begin(); itr != potentials_.end(); itr++)
    delete *itr;
}

int TreeCRF::ClassifyExample(const Example & ex)  const {
  double best_potential = -1E300;
  int best_class = -1;
  for (int iclass = 0; iclass < (int)potentials_.size(); ++iclass) {
    double this_potential = potentials_[iclass]->Score(ex);
    if (this_potential > best_potential) {
      best_potential = this_potential;
      best_class = iclass;
    }
  }
  return best_class;
}

void TreeCRF::GetExpScores(const Example & ex,
                           std::vector<double> * scores) const {
  for (int iclass = 0; iclass < (int) potentials_.size(); ++iclass)
    (*scores)[iclass] = exp(potentials_[iclass]->Score(ex));
}

void TreeCRF::GetScores(const Example & ex,
                        std::vector<double> * scores) const {
  for (int iclass = 0; iclass < (int)potentials_.size(); ++iclass)
    (*scores)[iclass] = potentials_[iclass]->Score(ex);
}

void TreeCRF::GetExampleProb(const Example & ex,
                             std::vector<double> * probs) const {
    GetExpScores(ex, probs);
    double normalizer = 0.0;
    for (int iclass = 0; iclass < (int)potentials_.size(); ++iclass)
      normalizer += (*probs)[iclass];

    if (normalizer == 0)
      std::cerr << "GetExampleProb dividing by zero" << std::endl;

    for (int iclass = 0; iclass < (int)potentials_.size(); ++iclass)
      (*probs)[iclass] /= normalizer;
}

double TreeCRF::GetClassScore(const Example & ex, int targetClass) const {
    return potentials_[targetClass]->Score(ex);
}

void TreeCRF::BoostML(const std::vector<DataSet *> & train_words,
                      ForwardBackwardEval & fb_eval) {
  // create a training set that samples the original data set.
  // This is to support stochastic tree boosting.
  std::vector<DataSet *> sample_train_words(params.maxSetSize);
  RandomSample(train_words, sample_train_words);

  int sample_size = (int)sample_train_words.size();

  // std::cout << "Forward-Backward Iteration" << std::endl;
  // perform a full forward-backward iteration for CRF training.
  std::cout << "Allocating trellis" << std::endl;
  std::vector<VNodePtrMatrix *> trellis_vec(sample_size);

  for (int iexample = 0; iexample < sample_size; ++iexample) {
    // std::cout << "Forward-Backward for example " << iexample << std::endl;
    // trellis_vec[iexample] =
    trellis_vec[iexample] = new VNodePtrMatrix(sample_train_words[iexample]->GetNumExamples());
    fb_eval.ForwardBackward(*(sample_train_words[iexample]), *this, trellis_vec[iexample]);
  }

  // std::cout << "Forward-backward complete, now generating RT training examples" << std::endl;
  for (int iclass = 0; iclass < sample_train_words[0]->num_classes(); ++iclass) {
    // std::cout << "Building data for class " << iclass << std::endl;
    DataSet new_data;
    new_data.set_num_features(sample_train_words[0]->num_features());
    new_data.set_num_classes(sample_train_words[0]->num_classes());

    SetTargetsML(sample_train_words, trellis_vec, iclass, &new_data);

    // std::cout << "Building tree for class " << iclass << std::endl;
    potentials_[iclass]->Train(new_data, params.maxLeaves);
    new_data.DeleteExamples();
  }

  // Reclaim the trellis_vec
  for (std::vector<VNodePtrMatrix *>::iterator itr = trellis_vec.begin();
       itr != trellis_vec.end(); itr++)
    fb_eval.DeleteTrellis(*itr);
}

void TreeCRF::BoostRSW(const std::vector<DataSet *> & train_words) {
  std::cout << "creating sample_train_words" << std::endl;
  // create a training set that samples the original data set.
  // This is to support stochastic tree boosting.
  std::vector<DataSet *> sample_train_words(params.maxSetSize);
  RandomSample(train_words, sample_train_words);

  DataSet new_data;
  new_data.set_num_features(sample_train_words[0]->num_features());
  new_data.set_num_classes(sample_train_words[0]->num_classes());

  for (int iexample = 0; iexample < (int)sample_train_words.size(); ++iexample)
    for (int jexample = 0; jexample < sample_train_words[iexample]->GetNumExamples(); ++jexample)
      new_data.AddExample(sample_train_words[iexample]->GetExample(jexample));

  // Allocate the "probability matrix" that stores the predicted
  // class probabilities for every training example
  std::vector< std::vector<double> > probs(new_data.GetNumExamples());
  // Fill this matrix
  GetProbMatrix(new_data, probs);

  // Ignore the (-1) class, since we never predict it.  It is only
  // used for input
  for (int iclass = 0; iclass < new_data.num_classes(); ++iclass) {
    // Compute regression tree targets for this class
    SetTargetsRSW(probs, iclass, &new_data);
    // Build the regression tree
    potentials_[iclass]->Train(new_data, params.initMaxLeaves);
  }
}

void TreeCRF::Print(std::ostream & str) const {
  str << "TreeCRF " << (int) potentials_.size();
  std::vector<RTPotentialFunction *>::const_iterator itr;
  for (itr = potentials_.begin(); itr != potentials_.end(); itr++)
    str << " " << **itr;
}

void TreeCRF::Read(std::istream & str) {
  std::string buf;
  str >> buf;
  if (buf != std::string("TreeCRF")) {
    std::cerr << "Found `" << buf
              << "' where `TreeCRF' was expected."
              << std::endl;
    std::abort();
  }
  int nclasses = 0;
  str >> nclasses;
  for (int iclass = 0; iclass < nclasses; ++iclass) {
    // read in for cached version
    RTPotentialFunction * pf = new RTPotentialFunction(CacheSize);
    str >> *pf;
    potentials_[iclass] = pf;
  }
}

void TreeCRF::RandomSample(const std::vector<DataSet *> & train_words,
                           std::vector<DataSet *> & sample_train_words) {
  std::cout << "creating sample_train_words" << std::endl;
  int sample_size = 0;
  for (int iexample = 0; iexample < (int)train_words.size(); iexample++) {
    if ((params.sampleRate == 1) ||
        ((double)std::rand()) / ((double)RAND_MAX) < params.sampleRate) {
      // This is a shallow copy (just DataSet *)
      sample_train_words[sample_size] = train_words[iexample];
      sample_size++;
    }
  }
  // std::cout << "resizing sampleTrainSet" << std::endl;
  sample_train_words.resize(sample_size);
}

void TreeCRF::SetTargetsML(const std::vector<DataSet *> & train_words,
                           const std::vector<VNodePtrMatrix *> & trellis_vec,
                           int target_class,
                           DataSet * dataset) {
  VNodePtrMatrix *trellis;
  VNodePtrVec *alpha_column, *beta_column;
  Example *alpha_ex, *beta_ex, *new_ex;

  int match, matched;
  int state_len;
  double normalizer;

  std::vector<int> true_state(k_val_ + 1, -1);
  int num_classes = dataset->num_classes();
  int recurrent_feature = dataset->num_features() - (num_classes + 1) * k_val_;

  // std::cout << "Setting targets..." << std::endl;
  for (int isequence = 0; isequence < (int) train_words.size(); ++isequence) {
    trellis = trellis_vec[isequence];

    int nwindows = train_words[isequence]->GetNumExamples();

    int beta_col = 0;
    beta_ex = train_words[isequence]->GetExample(beta_col);
    state_len = GetTrueState(train_words[isequence], beta_col, true_state);
    beta_column = (*trellis)[beta_col];

    // Finding the normalizer for our first 'column' in the matrix
    normalizer = 0.0;
    for (VecIter itr = beta_column->begin();
         itr != beta_column->end(); itr++)
      normalizer += (*itr)->alpha() * (*itr)->beta();

    matched = 0;
    for (VecIter itr = beta_column->begin();
         itr != beta_column->end(); itr++) {
      // Since we are checking the very beginning of the forward-backward
      // trellis, there is only one valid state feature to check for a match
      if ((*itr)->GetStateAt(0) == target_class) {
	match = (true_state[0] == target_class);
	// mark if the 'true' state appears among the possible nodes
	if (match) matched = 1;

	if (match) {
	  new_ex = new Example(*beta_ex);
	  new_ex->set_target(1 - (*itr)->alpha() * (*itr)->beta() / normalizer);
          dataset->AddExample(new_ex);
	} else {
	  new_ex = new Example(*beta_ex);
	  new_ex->set_category(target_class);
	  new_ex->set_target(- (*itr)->alpha() * (*itr)->beta() / normalizer);
	  dataset->AddExample(new_ex);
	}
      }
    }
    if (!(matched) && (beta_ex->category() == target_class)) {
      new_ex = new Example(*beta_ex);
      new_ex->set_target(1.0);
      dataset->AddExample(new_ex);
    }

    // At this point we have finished working on column 0 of the trellis, which
    // has no predecessor.  Now we process the rest of the columns.
    for (int alpha_col = 0; alpha_col < nwindows-1; ++alpha_col) {
      int beta_col = alpha_col + 1;

      alpha_ex = train_words[isequence]->GetExample(alpha_col);
      beta_ex = train_words[isequence]->GetExample(beta_col);

      state_len = GetTrueState(train_words[isequence], beta_col, true_state);
      matched = 0;

      // computing the normalizer by summing all possible (alpha * f_k * beta) paths between
      // the nodes in the alpha and beta coloumns
      normalizer = 0;
      alpha_column = (*trellis)[alpha_col];
      beta_column = (*trellis)[beta_col];

      for (VecIter beta_iter = beta_column->begin();
           beta_iter != beta_column->end(); beta_iter++) {
	int beta_class = (*beta_iter)->GetStateAt(0);
        for (VecIter alpha_iter = alpha_column->begin();
             alpha_iter != alpha_column->end(); alpha_iter++)
	  if ((*alpha_iter)->IsParent(**beta_iter)) {
	    normalizer +=
              (*alpha_iter)->alpha() *
              (*alpha_iter)->GetOutProbAt(beta_class) *
              (*beta_iter)->beta() ;
          }
      }

      for (VecIter beta_iter = beta_column->begin();
           beta_iter != beta_column->end(); beta_iter++) {
	int beta_class = (*beta_iter)->GetStateAt(0);
	if (beta_class == target_class) {
	  for (VecIter alpha_iter = alpha_column->begin();
               alpha_iter != alpha_column->end(); alpha_iter++) {

	    if ((*alpha_iter)->IsParent(**beta_iter)) {
	      match = SearchForMatch(true_state, beta_class, *alpha_iter, state_len);

	      // mark if the 'true' state appears among the probable nodes
	      if (match) matched = 1;
	      if (match) {
		new_ex = new Example(*beta_ex);
		new_ex->set_target(1-
                                   (*alpha_iter)->alpha() *
                                   (*alpha_iter)->GetOutProbAt(beta_class) *
                                   (*beta_iter)->beta() / normalizer);
		dataset->AddExample(new_ex);
	      } else {
		new_ex = new Example(*beta_ex);
                new_ex->set_target(-1 *
                                   (*alpha_iter)->alpha() *
                                   (*alpha_iter)->GetOutProbAt(beta_class) *
                                   (*beta_iter)->beta() / normalizer);
		SetNewExample(*alpha_iter, beta_class, recurrent_feature,
                              num_classes, new_ex);
		new_ex->set_category(target_class);
		dataset->AddExample(new_ex);
	      }
	    }
	  }
	}
      }

      if (!(matched) && (beta_ex->category() == target_class)) {
	new_ex = new Example(*beta_ex);
	new_ex->set_target(1.0);
	dataset->AddExample(new_ex);
      }
    }
  }

  if (params.trim) {
    for (int iexample = 0; iexample < dataset->GetNumExamples(); ++iexample) {
      new_ex = dataset->GetExample(iexample);
      //Trimming!
      if (fabs(new_ex->target())*(1-fabs(new_ex->target())) < params.trim_val)
        new_ex->set_skip(true);
    }
  }
}

int TreeCRF::GetTrueState(DataSet * data,
                          int offset,
                          std::vector<int> & true_state) {
  int state_len = std::min(k_val_ + 1, offset + 1);
  for (int i = 0; i < state_len; ++i)
    true_state[i] = data->GetExample(offset - i)->category();
  return state_len;
}

int TreeCRF::SearchForMatch(const std::vector<int> & true_state,
                            int current_class,
                            ViterbiNode * test_node,
                            int state_len) {
  if (true_state[0] != current_class) return 0;
  for (int x = 1; x < state_len; ++x)
    if (true_state[x] != test_node->GetStateAt(x-1)) return 0;
  return 1;
}

void TreeCRF::SetNewExample(ViterbiNode * alpha_node,
                            int beta_class,
                            int recurrent_feature,
                            int num_classes,
                            Example * new_ex) {
  new_ex->set_category(beta_class);
  int upper_bound = std::min(k_val_, alpha_node->trellis_index() + 1);
  int loc = new_ex->GetSize() - k_val_;
  for (int i = 0; i < upper_bound; ++i) {
    new_ex->SetFeatureAt(loc, recurrent_feature + alpha_node->GetStateAt(i) + 1);
    loc++;
    recurrent_feature += num_classes + 1;
  }
}

void TreeCRF::GetProbMatrix(const DataSet & data,
                            std::vector<std::vector<double> > & probs) {
  std::vector<double> col_probs(data.num_classes());
  for (int x = 0; x < data.GetNumExamples(); ++x) {
    GetExampleProb(*data.GetExample(x), &col_probs);
    probs[x] = col_probs;
  }
}

void TreeCRF::SetTargetsRSW(const std::vector<std::vector<double> > & probs,
                            int true_class, DataSet * data) {
  for (int i = 0; i < data->GetNumExamples(); ++i) {
    Example* ex = data->GetExample(i);
    ex->set_target((ex->category() == true_class) - probs[i][true_class]);
    if (params.trim)
      ex->set_skip(fabs(ex->target())*(1-fabs(ex->target())) < 0.001);
    else
      ex->set_skip(false);
  }
}

void operator << (std::ostream & str, const TreeCRF & pfc) {
  pfc.Print(str);
}

void operator >> (std::istream & str, TreeCRF & pfc) {
  pfc.Read(str);
}

} // namespace treecrf

//////////////////////////////////////////////////////////////////////
