// File: treecrf.h
//
// Time-stamp: <2008-07-18 01:38:05 haog>
// Author: Guohua Hao <haog@eecs.oregonstate.edu>
//
// Copyright (C) 2008 Guohua Hao
// All Rights Reserved.
//
// Description: class for tree CRF model
//
//////////////////////////////////////////////////////////////////////

#ifndef _TREECRF_H_
#define _TREECRF_H_

#include <vector>
#include "potential_function.h"

namespace treecrf {

class DataSet;
class Example;
class ForwardBackwardEval;
class ViterbiNode;

// TreeCRF model based on RTPotentialFunction.  It stores one potential function for each class.
class TreeCRF {
 public:
  typedef std::vector<ViterbiNode *> VNodePtrVec;
  typedef VNodePtrVec::iterator VecIter;
  typedef std::vector<VNodePtrVec *> VNodePtrMatrix;
  typedef VNodePtrMatrix::iterator MatrixIter;

  // Constructors
  TreeCRF(): k_val_(1), potentials_(0) {};

  TreeCRF(int k_val, int num_classes, int cache_size):
    k_val_(k_val), potentials_(num_classes) {
    for (int i = 0; i < num_classes; ++i)
      potentials_[i] = new RTPotentialFunction(cache_size);
  };

  ~TreeCRF();

  // Return predicted class for this example (class with highest potential value)
  int ClassifyExample(const Example & ex) const;

  // Return the unnormalized exponentiated potential values
  void GetExpScores(const Example & ex, std::vector<double> * scores) const;

  // Return the unnormalized potential values
  void GetScores(const Example & ex, std::vector<double> * scores) const;

  // Return vector of class probabilities
  void GetExampleProb(const Example & ex, std::vector<double> * probs) const;

  // Compute the (unexponentiated) potential score for a specific class targetClass
  double GetClassScore(const Example & ex, int targetClass) const;

  // Update the classifier by gradient tree boosting
  // based on maximum likelihood (ML)
  void BoostML(const std::vector<DataSet *> & train_words,
               ForwardBackwardEval & fb_eval);

  // Update the classifier by gradient tree boosting based on
  // recurrent sliding window (RSW)
  void BoostRSW(const std::vector<DataSet *> & train_words);

  // IO
  void Print(std::ostream & str) const;
  void Read(std::istream & str);

 private:
  // create a training set that samples the original data set.
  // This is to support stochastic tree boosting.
  void RandomSample(const std::vector<DataSet *> & train_words,
                    std::vector<DataSet *> & sample_train_words);

  // Set training targets based on results of forward-backward applied to the trellis
  void SetTargetsML(const std::vector<DataSet *> & train_words,
                    const std::vector<VNodePtrMatrix *> & trellis_vec,
                    int target_class,
                    DataSet * dataset);

  // Get the true states of the positions in the given data. The
  // positions start from offset and go backwards.
  int GetTrueState(DataSet * data,
                   int offset,
                   std::vector<int> & true_state);

  // Check if the current transition exists in the original sequence
  int SearchForMatch(const std::vector<int> & true_state,
                     int current_class,
                     ViterbiNode * test_node,
                     int state_len);

  // Set the recurrent features for the new example based on the alpha
  // and beta nodes
  void SetNewExample(ViterbiNode * alpha_node,
                     int beta_class,
                     int recurrent_feature,
                     int num_classes,
                     Example * new_ex);

  // Get the probability matrix of P(y | x_t) for each x_t in the dataset and
  // each possible class label y.
  void GetProbMatrix(const DataSet & data,
                     std::vector<std::vector<double> > & probs);

  // Set training targets based on recurrent sliding window (RSW) method
  void SetTargetsRSW(const std::vector<std::vector<double> > & probs,
                     int true_class, DataSet * data);

  // Data Fields
  int k_val_;
  std::vector<RTPotentialFunction *> potentials_; // indexed by class number
};

void operator << (std::ostream & str, const TreeCRF & crf);
void operator >> (std::istream & str, TreeCRF & crf);

} // namespace treecrf

#endif  // _TREECRF_H_
//////////////////////////////////////////////////////////////////////
