// File: viterbi.cc
//
// Time-stamp: <2008-07-17 18:48:41 haog>
// Author: Guohua Hao <haog@eecs.oregonstate.edu>
//
// Copyright (C) 2008 Guohua Hao
// All Rights Reserved.
//
// Description: implementation of ViterbiEval class
//
//////////////////////////////////////////////////////////////////////

#include <vector>
#include <algorithm>
#include "dataset.h"
#include "rt_main.h"
#include "treecrf.h"
#include "viterbi.h"
#include "viterbi_node.h"

namespace treecrf {

void ViterbiEval::Predict(const DataSet & dataset,
                          const TreeCRF & crf,
                          int* predict_classes) {
  VNodePtrList parents;
  int nwindows = dataset.GetNumExamples(); // length of sequence
  int num_features = dataset.num_features();
  int num_classes = dataset.num_classes();

  VNodePtrMatrix * trellis = new VNodePtrMatrix(nwindows);

  // Example for the first step
  Example * ex = dataset.GetExample(0);
  // the first node of the forward pass
  ViterbiNode * start = new ViterbiNode(k_val_, num_classes);
  start->ComputeOutputProbScore(*ex, crf);
  parents.push_back(start);

  // best_nodes_ is empty, parents contains just the start state
  GrowNextHeap(parents, num_classes);
  (*trellis)[0] = new VNodePtrVec;
  BuildNodeMatrix((*trellis)[0]);

  for (int x = 1; x < nwindows; ++x) {
    // cout << "Node " << x << " of forward pass" << endl;
    ex = new Example(*dataset.GetExample(x));
    BuildParents((*trellis)[x-1], ex, crf, num_features,
                 num_classes + 1, parents);
    delete ex;
    GrowNextHeap(parents, num_classes);
    (*trellis)[x] = new VNodePtrVec;
    BuildNodeMatrix((*trellis)[x]);
  }

  // std::cout << "Backtrace phase!" << std::endl;
  VecIter max_itr = std::min_element((*trellis)[nwindows-1]->begin(),
                                     (*trellis)[nwindows-1]->end(),
                                     ViterbiNodeComparator());
  ViterbiNode * btrace = *max_itr;
  while (btrace->trellis_index() > -1) {
    predict_classes[btrace->trellis_index()] = btrace->GetStateAt(0);
    btrace = btrace->previous_node();
  }

  delete start;
  DeleteTrellis(trellis);
}

void ViterbiEval::GrowNextHeap(VNodePtrList & parents, int num_classes) {
  // Clear heap
  while(!best_nodes_.empty())
    best_nodes_.pop();

  VNodePtrList sub_parents;
  while (!parents.empty()) {
    BuildSubParents(parents, sub_parents);
    for (int x = 0; x < num_classes; ++x) {
      ViterbiNode * temp = new ViterbiNode(k_val_, num_classes);
      temp->SetOldStates(*(sub_parents.front()));
      temp->SetCurentStates(x);
      temp->IncTrellisIndex(sub_parents.front()->trellis_index());
      temp->UpdateProb(sub_parents);
      UpdateHeap(temp);
    }
  }
}

void ViterbiEval::BuildSubParents(VNodePtrList & parents,
                                  VNodePtrList & sub_parents) {
  // throw anything out that might be left over
  sub_parents.clear();
  if (parents.empty()) {
    std::cout << "Warning: Trying to make the sub_parent list with empty parent list!"
              << std::endl;
    return;
  }

  VNodePtrList temp_parents;
  ViterbiNode * top = parents.front();
  while (!parents.empty()) {
    ViterbiNode * current = parents.front();
    if (top->Match(*current))
      sub_parents.push_back(current);
    else
      temp_parents.push_back(current);
    parents.pop_front();
  }

  while (!temp_parents.empty()) {
    parents.push_back(temp_parents.front());
    temp_parents.pop_front();
  }
}

void ViterbiEval::UpdateHeap(ViterbiNode * current) {
  if ((int)best_nodes_.size() < beam_breadth_) {
    best_nodes_.push(current);
  } else if ((int) best_nodes_.size() > beam_breadth_) {
    std::cout << "Error: the priority queue is larger than the max breadth"
              << std::endl;
  } else {
    if (*best_nodes_.top() < *current) {
      heapshrink++;
      ViterbiNode * top_node = best_nodes_.top();
      best_nodes_.pop();
      best_nodes_.push(current);
      delete top_node;
    } else {
      delete current;
    }
  }
}

void ViterbiEval::BuildNodeMatrix(VNodePtrVec * current_column) {
  while (!best_nodes_.empty()) {
    current_column->push_back(best_nodes_.top());
    best_nodes_.pop();
  }
}

void ViterbiEval::BuildParents(VNodePtrVec * current_column,
                               Example * ex,
                               const TreeCRF & crf,
                               int num_features,
                               int num_classes,
                               VNodePtrList & parents) {
  parents.clear();
  int loc = ex->GetSize() - k_val_;
  int offset = num_features - (num_classes * k_val_);
  int upper_bound = std::min((*current_column)[0]->trellis_index() + 1, k_val_);

  for (VecIter itr = current_column->begin();
       itr != current_column->end(); itr++) {
    ViterbiNode * current_node = (*itr);
    int recurrent_feature = offset;
    int current_loc = loc;
    for (int y = 0; y < upper_bound; ++y) {
      ex->SetFeatureAt(current_loc, recurrent_feature + current_node->GetStateAt(y) + 1);
      current_loc++;
      recurrent_feature += num_classes;
    }
    current_node->ComputeOutputProbScore(*ex, crf);
    parents.push_back(current_node);
  }
}

void ViterbiEval::DeleteTrellis(VNodePtrMatrix * trellis) {
  if (trellis == 0) return;
  for (MatrixIter col_itr = trellis->begin();
       col_itr != trellis->end(); col_itr++) {
    for(VecIter row_itr = (*col_itr)->begin();
        row_itr != (*col_itr)->end(); row_itr++) {
      delete *row_itr;
    }
    delete *col_itr;
  }
  delete trellis;
  trellis = 0;
}

}; // namespace treecrf

//////////////////////////////////////////////////////////////////////
