// File: viterbi.h
//
// Time-stamp: <2008-07-17 18:49:32 haog>
// Author: Guohua Hao <haog@eecs.oregonstate.edu>
//
// Copyright (C) 2008 Guohua Hao
// All Rights Reserved.
//
// Description: ViterbiEval class, which is a viterbi algorithm wrapper
//
//////////////////////////////////////////////////////////////////////

#ifndef _VITERBI_H_
#define _VITERBI_H_

#include <list>
#include <queue>
#include <vector>
#include "viterbi_node.h"

namespace treecrf {

class ViterbiEval {
 public:
  typedef std::vector<ViterbiNode *> VNodePtrVec;
  typedef VNodePtrVec::iterator VecIter;
  typedef std::vector<VNodePtrVec *> VNodePtrMatrix;
  typedef VNodePtrMatrix::iterator MatrixIter;
  typedef std::list<ViterbiNode *> VNodePtrList;

  // Constructors
  ViterbiEval(): k_val_(1), beam_breadth_(10) {};

  ViterbiEval(int k_val, int beam_breadth):
    k_val_(k_val), beam_breadth_(beam_breadth) {};

  // Predict the label sequence of a given sequence dataset
  void Predict(const DataSet & dataset,
               const TreeCRF & crf,
               int* predict_classes);

 private:
  // Generate the next heap from the parents
  void GrowNextHeap(VNodePtrList & parents, int num_classes);

  // Builds a sublist of parents, where each node has the same possible range
  // for their children
  void BuildSubParents(VNodePtrList & parents, VNodePtrList & sub_parents);

  // Potentially put the current node into the PQ in the beam search
  void UpdateHeap(ViterbiNode * current);

  // Build the next column of the trellin based on the nodes
  // in the PQ, i.e., the beam
  void BuildNodeMatrix(VNodePtrVec * current_column);

  // Assign each node in the current column their scores, and form a list of
  // parent nodes
  void BuildParents(VNodePtrVec * current_column,
                    Example * ex,
                    const TreeCRF & crf,
                    int num_features,
                    int num_classes,
                    VNodePtrList & parents);

  // Reclaim the memories for all nodes in the trellis and the trellis itself.
  // Remember to call this function after you are done with the trellis.
  void DeleteTrellis(VNodePtrMatrix * trellis);

  // Data Fields
  int k_val_; // the order of markov chain
  int beam_breadth_; // the breadth of beam
  std::priority_queue<ViterbiNode *, std::vector<ViterbiNode *>, ViterbiNodeComparator> best_nodes_;
};

}; // namespace treecrf

#endif  // _VITERBI_H_
//////////////////////////////////////////////////////////////////////
