// File: viterbi_node.cc
//
// Time-stamp: <2008-07-17 16:50:56 haog>
// Author: Guohua Hao <haog@eecs.oregonstate.edu>
//
// Copyright (C) 2008 Guohua Hao
// All Rights Reserved.
//
// Description: implementation of the class ViterbiNode
//
//////////////////////////////////////////////////////////////////////

#include <list>
#include "treecrf.h"
#include "viterbi_node.h"

namespace treecrf {

bool ViterbiNode::operator < (const ViterbiNode& a) const {
  if (prob_ < a.prob_) return true;
  return false;
}

bool ViterbiNode::Match(const ViterbiNode & a) const {
  if ((int)states_.size() == 1) return true;
  for (int x = 0; x < (int)states_.size() - 1; ++x)
    if (states_[x] != a.states_[x]) return false;
  return true;
}

bool ViterbiNode::IsParent(const ViterbiNode & child) const {
  for (int x = 1; x < (int)child.states_.size(); ++x)
    if (states_[x-1] != child.states_[x]) return false;
  return true;
}

void ViterbiNode::SetOldStates (const ViterbiNode & a) {
  if ((int)states_.size() == 1) return;
  for (int x = 1; x < (int)states_.size(); ++x)
    states_[x] = a.states_[x-1];
}

void ViterbiNode::ComputeOutputProb(const Example & ex, const TreeCRF & crf) {
  crf.GetExpScores(ex, &out_probs_);
}

void ViterbiNode::ComputeOutputProbScore(const Example & ex, const TreeCRF & crf) {
  crf.GetScores(ex, &out_probs_);
}

void ViterbiNode::UpdateAlpha(const std::list<ViterbiNode* > & parents) {
  int new_state = states_[0];
  alpha_ = 0.0;
  std::list<ViterbiNode *>::const_iterator itr;
  for (itr = parents.begin(); itr != parents.end(); itr++)
    alpha_ += (*itr)->alpha_ * (*itr)->out_probs_[new_state];

  //we set the prob since it is used to prioritize the heap
  prob_ = alpha_;
}

void ViterbiNode::UpdateProb(const std::list<ViterbiNode* > & parents) {
  int new_state = states_[0];
  prob_ = -1e100;
  std::list<ViterbiNode *>::const_iterator itr;
  for (itr = parents.begin(); itr != parents.end(); itr++) {
    double temp = (*itr)->prob_ + (*itr)->out_probs_[new_state];
    if (temp > prob_) {
      prob_ = temp;
      previous_node_ = *itr;
    }
  }
}

void ViterbiNode::UpdateBeta(const std::vector<ViterbiNode* > & children) {
  beta_ = 0.0;
  std::vector<ViterbiNode *>::const_iterator itr;
  for (itr = children.begin(); itr != children.end(); itr++) {
    if (IsParent(**itr)) {
      int child_state = (*itr)->GetStateAt(0);
      beta_ += out_probs_[child_state] * (*itr)->beta_;
    }
  }
}

} // namespace treecrf

//////////////////////////////////////////////////////////////////////
