// File: viterbi_node.h
//
// Time-stamp: <2008-07-17 17:36:54 haog>
// Author: Guohua Hao <haog@eecs.oregonstate.edu>
//
// Copyright (C) 2008 Guohua Hao
// All Rights Reserved.
//
// Description: class for each node in the FB/Viterbi decoding
//
//////////////////////////////////////////////////////////////////////

#ifndef _VITERBI_NODE_H_
#define _VITERBI_NODE_H_

#include <vector>

namespace treecrf {

class Example;
class TreeCRF;

class ViterbiNode {
 public:
  // Constructors
  ViterbiNode(int k_val, int num_classes):
    alpha_(1.0), beta_(1.0), prob_(0.0), trellis_index_(-1),
    previous_node_(0), states_(k_val, -1),
    out_probs_(num_classes, 1.0) {};

  // Compare the 'prob_' field.
  bool operator < (const ViterbiNode & a) const;

  // Find whether two ViterbiNodes have the same successors
  bool Match(const ViterbiNode & a) const;

  // Check if the current node is the parent node
  bool IsParent(const ViterbiNode & a) const;

  // Set the recurrent states from the given node, which is usually a parent node
  void SetOldStates(const ViterbiNode & a);

  void SetCurentStates(int current_state) { states_[0] = current_state; };

  void IncTrellisIndex(int index) { trellis_index_ = index + 1; };

  // Compute the output exponentiated scores for a given examaple and crf
  void ComputeOutputProb(const Example & ex, const TreeCRF & crf);

  // Compute the output score for a given examaple and crf
  void ComputeOutputProbScore(const Example & ex, const TreeCRF & pfc);

  // Update the alpha value based on its parent nodes.
  // Used in forward backward algorithm.
  void UpdateAlpha(const std::list<ViterbiNode *> & parents);

  // Update the prob value based on its parent nodes. Used in viterbi algorithm
  void UpdateProb(const std::list<ViterbiNode *> & parents);

  // Update the beta value based on its children nodes.
  // Used in forward backward algorithm.
  void UpdateBeta(const std::vector<ViterbiNode *> & children);

  void DivideAlpha(double normalizer) { alpha_ /= normalizer; };
  void DivideBeta(double normalizer) { beta_ /= normalizer; };

  // Getters
  double alpha() const { return alpha_; };
  double beta() const { return beta_; };
  double prob() const { return prob_; };
  int trellis_index() const { return trellis_index_; };
  ViterbiNode * previous_node() const { return previous_node_; };
  int GetStateAt(int index) const { return states_[index]; };
  double GetOutProbAt(int iclass) const { return out_probs_[iclass]; };

 private:
  //For forward backward calcs
  double alpha_;
  double beta_;
  double prob_;
  int trellis_index_;
  ViterbiNode * previous_node_;
  // Each viterbi node stores the values of y_t, y_{t-1}, ... y_{t-npad +1},
  // 0 is current.
  std::vector<int> states_;
  // P(y_{t+1} | this node)
  std::vector<double> out_probs_;
};

class ViterbiNodeComparator {
 public:
  bool operator () (ViterbiNode * a, ViterbiNode * b) {
    if (a->prob() > b->prob()) return true;
    return false;
  }
};

} // namespace treecrf

#endif // _VITERBI_NODE_H_
//////////////////////////////////////////////////////////////////////
